from pucktrick.utils import *
from pucktrick.noisy import *
import pandas as pd
import numpy as np

def missingNew(train_df,column,percentage):
  extracted_list=sampleList(percentage,len(train_df[column]))
  noise_df= train_df.copy()
  for i, value in enumerate(extracted_list):
    noise_df.loc[i, column] = np.nan 
  return noise_df

def missingExtended(original_df,train_df,column,percentage):
    noise_df= train_df.copy()
    noise_df['id1'] = range(len(noise_df))
    new_df,newPercentage=generateSubdf(original_df, noise_df,column,percentage)
    if newPercentage==0:
      return train_df
    modified_df= missingNew(new_df,column,newPercentage)
    i=0
    for index in modified_df['id1']:
        if pd.isnull(modified_df[modified_df['id1']==index][column]).any() and pd.notnull(noise_df[noise_df['id1']==index][column]).any():
            i += 1
            indexer=noise_df[noise_df['id1']==index].index
            noise_df.loc[indexer,column] = np.nan   
    noise_df = noise_df.drop('id1', axis=1)
    return noise_df





from pucktrick.utils import *
from pucktrick.noisy import *
import pandas as pd

def outlierContinuosNew3Sigma(train_df,column,percentage):
  extracted_list=sampleList(percentage,len(train_df[column]))
  noise_df= train_df.copy()
  mean = np.mean(noise_df[column])
  std_dev = np.std(noise_df[column])
  upper_bound=mean + 3 * std_dev
  lower_bound=mean -3 * std_dev
  new_lower_limit = mean - 4 * std_dev
  new_upper_limit = mean + 4 * std_dev

  for i, value in enumerate(extracted_list):
     if np.random.rand() > 0.5:
        noise_df.loc[i, column] = generate_random_value(upper_bound, new_upper_limit)
     else:
        noise_df.loc[i, column] = generate_random_value(new_lower_limit, lower_bound)
        
  return noise_df

def outlierContinuosExtended3Sigma(original_df, train_df,column,percentage):
  noise_df= train_df.copy()
  noise_df['id1'] = range(len(noise_df))
  new_df,newPercentage=generateSubdf(original_df, noise_df,column,percentage)
  if newPercentage==0:
      return train_df
  modified_df= outlierContinuosNew3Sigma(new_df,column,newPercentage)
  noise_df=mergeDataframe(noise_df,modified_df)
  return noise_df

def outlierDiscreteNew3Sigma(train_df,column,percentage):
  extracted_list=sampleList(percentage,len(train_df[column]))
  noise_df= train_df.copy()
  mean = np.mean(noise_df[column])
  std_dev = np.std(noise_df[column])
  upper_bound=mean + 3 * std_dev
  lower_bound=mean -3 * std_dev
  new_upper_limit = mean + 4 * std_dev
  new_lower_limit = mean - 4 * std_dev
  if upper_bound==0:
     upper_bound=1
  if new_upper_limit==upper_bound:
     new_upper_limit=5*upper_bound
  
  if lower_bound==0:
     lower_bound=-1
  if new_lower_limit==lower_bound and lower_bound<0:
     new_lower_limit=5*lower_bound
  elif lower_bound>0:
     new_lower_limit=-5*lower_bound
  if upper_bound>new_upper_limit:
      tmp=new_upper_limit
      new_upper_limit=upper_bound
      upper_bound=tmp
  if new_lower_limit>lower_bound:
     tmp=lower_bound
     lower_bound=new_lower_limit
     new_lower_limit=tmp
  for i, value in enumerate(extracted_list):
     if np.random.rand() > 0.5:
        noise_df.loc[i, column] = generate_random_value_discrete(upper_bound, new_upper_limit)
     else:
        noise_df.loc[i, column] = generate_random_value_discrete(new_lower_limit, lower_bound)
        
  return noise_df

def outlierDiscreteExtended3Sigma(original_df, train_df,column,percentage):
  noise_df= train_df.copy()
  noise_df['id1'] = range(len(noise_df))
  new_df,newPercentage=generateSubdf(original_df, noise_df,column,percentage)
  if newPercentage==0:
      return train_df
  modified_df= outlierDiscreteNew3Sigma(new_df,column,newPercentage)
  noise_df=mergeDataframe(noise_df,modified_df)
  return noise_df

def outlierCategoricalIntegerNew(train_df,column,percentage):
  extracted_list=sampleList(percentage,len(train_df[column]))
  noise_df= train_df.copy()
  max_value = noise_df[column].max()
  min_value=noise_df[column].min()
  if min_value==0:
      min_value=-1
      new_minValue=-5
  elif min_value<0:
      new_minValue=5*min_value
  else:
      new_minValue=-5+min_value
  if max_value==0:
      max_value=1
      new_maxValue=5
  else:
      new_maxValue=2*max_value
  for i, value in enumerate(extracted_list):
     if np.random.rand() > 0.5:
      noise_df.loc[i, column] = np.random.randint(max_value, new_maxValue)
     else:
      noise_df.loc[i, column] = np.random.randint(new_minValue, min_value)
        
  return noise_df

def outliercategoricalIntegerExtended(original_df, train_df,column,percentage):
  noise_df= train_df.copy()
  noise_df['id1'] = range(len(noise_df))
  new_df,newPercentage=generateSubdf(original_df, noise_df,column,percentage)
  if newPercentage==0:
      return train_df
  modified_df= outlierCategoricalIntegerNew(new_df,column,newPercentage)
  noise_df=mergeDataframe(noise_df,modified_df)
  return noise_df

def outlierCategoricalStringNew(train_df,column,percentage):
  extracted_list=sampleList(percentage,len(train_df[column]))
  noise_df= train_df.copy()
  for i, value in enumerate(extracted_list):
    noise_df.loc[i, column] = "puck was here"
  return noise_df

def outliercategoricalStringExtended(original_df, train_df,column,percentage):
  noise_df= train_df.copy()
  noise_df['id1'] = range(len(noise_df))
  new_df,newPercentage=generateSubdf(original_df, noise_df,column,percentage)
  if newPercentage==0:
      return train_df
  modified_df= outlierCategoricalStringNew(new_df,column,newPercentage)
  noise_df=mergeDataframe(noise_df,modified_df)
  return noise_df



import random
import pandas as pd
import numpy as np

def create_fake_table(num_rows=1000):
  # Generate data for each column
  f1 = np.random.uniform(-100, 100, num_rows)  # Continuous values between -100 and 100
  f2 = np.random.randint(-100, 101, num_rows)  # Discrete values between -100 and 100
  f3 = np.random.choice(['apple', 'banana', 'cherry'], num_rows)  # String values
  f4 = np.random.choice(['apple', 'banana', 'cherry'], num_rows)  # String values
  f5 = np.random.choice([0, 1], num_rows)  # Binary values (0 or 1)
  target = np.random.choice([0, 1], num_rows)  # Binary target (0 or 1)

  # Create the DataFrame
  df = pd.DataFrame({
    'f1': f1,
    'f2': f2,
    'f3': f3,
    'f4': f4,
    'f5': f5,
    'target': target
  })
  return df

def sampleList(percentage, maxValueList ):
    values = int(maxValueList * (percentage))
    extracted_List= random.sample(range(1, maxValueList), values)
    return extracted_List

def generateSubdf(original_df, train_df,column,percentage ):
    rowsToChange=len(original_df[column])*percentage
    dif = original_df[column] != train_df[column]
    diff_number = dif.sum()
    new_rowsToChange=rowsToChange-diff_number
    if new_rowsToChange<=0:
      newPercentage=0
      return train_df,newPercentage
    noise_df= train_df.copy()
    noise_df['id1'] = range(len(noise_df))
    or_df=original_df.copy()
    or_df['id1'] = range(len(or_df))
    mask = or_df[column] == noise_df[column]
    new_df = noise_df[mask]
    new_df = new_df.reset_index(drop=True)
    newPercentage=new_rowsToChange/len(new_df)
    return new_df,newPercentage

def mergeDataframe(noise_df,modified_df):

    merged_df = noise_df.merge(modified_df, on='id1', how='left', suffixes=('_df1', '_df2'))
    new_array = [string for string in noise_df.columns if string != 'id1']
    for col in new_array:
      noise_df[col] = merged_df[col + '_df2'].fillna(merged_df[col + '_df1'])
    noise_df = noise_df.drop('id1', axis=1)
    return noise_df

def generate_random_value(lower, upper):
        return np.random.uniform(lower, upper)

def generate_random_value_discrete(lower, upper):
        return np.random.randint(lower, upper)



from pucktrick.utils import *
import pandas as pd


def duplicateAllNew(train_df,percentage):
  rowsToChange=len(train_df)*percentage
  old_duplicated=train_df.duplicated().sum()
  new_rowsToChange=rowsToChange-old_duplicated
  if new_rowsToChange<=0:    
      return train_df
  percentage=new_rowsToChange/len(train_df)
  noise_df= train_df.copy()
  df_len=len(noise_df)
  extracted_list=sampleList(percentage,df_len)
  new_lines=int(df_len*percentage)
  for i in range(new_lines):
    tmpdf=train_df.loc[generate_random_value_discrete(0,len(extracted_list))].copy()
    noise_df.loc[len(noise_df)] = tmpdf         
  return noise_df

def duplicateAllExtended(original_df, train_df,percentage):
  rowsToChange=len(original_df)*percentage
  old_duplicated=original_df.duplicated().sum()
  new_duplicated=train_df.duplicated().sum()
  diff=new_duplicated-old_duplicated
  new_rowsToChange=rowsToChange-diff
  if new_rowsToChange<=0:    
      return train_df
  new_percentage=new_rowsToChange/len(original_df)
  noise_df= train_df.copy()
  extracted_list=sampleList(new_percentage,len(original_df))
  new_lines=int(len(original_df)*new_percentage)
  for i in range(new_lines):
    tmpdf=train_df.loc[generate_random_value_discrete(0,len(extracted_list))].copy()
    noise_df.loc[len(noise_df)] = tmpdf         
  return noise_df

def duplicateClassNew(train_df,target, value,percentage):
  target_df=train_df[train_df[target] ==value]
  target_df = target_df.reset_index(drop=True)
  rowsToChange=int(len(target_df)*percentage)
  old_duplicated=target_df.duplicated().sum()
  new_rowsToChange=rowsToChange-old_duplicated
  if new_rowsToChange<=0:    
      return train_df
  percentage=new_rowsToChange/len(target_df)
  noise_df= train_df.copy()
  df_len=len(target_df)
  extracted_list=sampleList(percentage,df_len) 
  for i in range(new_rowsToChange):
    row=generate_random_value_discrete(0,len(extracted_list))
    tmpdf=target_df.loc[row].copy()
    noise_df.loc[len(noise_df)] = tmpdf         
  return noise_df

def duplicateClassExtended(original_df, train_df,target, value,percentage):
  origin_target_df=original_df[original_df[target] ==value]
  target_df=train_df[train_df[target] ==value]
  target_df = target_df.reset_index(drop=True)
  origin_target_df = origin_target_df.reset_index(drop=True)
  rowsToChange=len(origin_target_df)*percentage
  old_duplicated=origin_target_df.duplicated().sum()
  new_duplicated=target_df.duplicated().sum()
  diff=new_duplicated-old_duplicated
  new_rowsToChange=int(rowsToChange-diff)
  if new_rowsToChange<=0:    
      return train_df
  percentage=new_rowsToChange/len(origin_target_df)
  noise_df= train_df.copy()
  df_len=len(origin_target_df)
  extracted_list=sampleList(percentage,df_len) 
  for i in range(new_rowsToChange):
    row=generate_random_value_discrete(0,len(extracted_list))
    tmpdf=origin_target_df.loc[row].copy()
    noise_df.loc[len(noise_df)] = tmpdf         
  return noise_df


import unittest
from pucktrick.outliers import *
from pucktrick.utils import *
import pandas as pd
import numpy as np
from pandas.testing import assert_frame_equal

class Test_outlies(unittest.TestCase):
  

  def test_outliersCategoricalString(self):
        percentage=0.5
        num_rows = 1000
        fake_df=create_fake_table(num_rows)
        column='f3'
        noise_df=outlierCategoricalStringNew(fake_df,column,percentage)
        dif = fake_df[column] != noise_df[column]
        diff_number = dif.sum()
        percentage=0.7
        noise_df=outliercategoricalStringExtended(fake_df,noise_df,column,percentage)
        dif = fake_df[column] != noise_df[column]
        diff_number = dif.sum()
        diff_p=diff_number/num_rows
        self.assertEqual(diff_p, percentage)
  
  def test_outliersCategoricalInt(self):
        percentage=0.5
        num_rows = 1000
        fake_df=create_fake_table(num_rows)
        column='f2'
        noise_df=outlierCategoricalIntegerNew(fake_df,column,percentage)
        dif = fake_df[column] != noise_df[column]
        diff_number = dif.sum()
        percentage=0.7
        noise_df=outliercategoricalIntegerExtended(fake_df,noise_df,column,percentage)
        dif = fake_df[column] != noise_df[column]
        diff_number = dif.sum()
        diff_p=diff_number/num_rows
        self.assertEqual(diff_p, percentage)
  
  def test_outliersdiscrete(self):
        percentage=0.5
        num_rows = 1000
        fake_df=create_fake_table(num_rows)
        column='f2'
        noise_df=outlierDiscreteNew3Sigma(fake_df,column,percentage)
        dif = fake_df[column] != noise_df[column]
        diff_number = dif.sum()
        percentage=0.7
        noise_df=outlierDiscreteExtended3Sigma(fake_df,noise_df,column,percentage)
        dif = fake_df[column] != noise_df[column]
        diff_number = dif.sum()
        diff_p=diff_number/num_rows
        # Verifica che il numero di modifiche sia corretto
        self.assertEqual(diff_p, percentage)

  def test_outlierscontinous(self):
        percentage=0.5
        num_r = 1000
        fake_df=create_fake_table(num_r)
        #fake_df=pd.read_csv("age_predictions_cleaned-3.csv", sep=";")
        num_rows=len(fake_df)
        column='f1'
        noise_df=outlierContinuosNew3Sigma(fake_df,column,percentage)
        dif = fake_df[column] != noise_df[column]
        diff_number = float(dif.sum())
        percentage=0.7
        noise_df=outlierContinuosExtended3Sigma(fake_df,noise_df,column,percentage)
        dif = fake_df[column] != noise_df[column]
        diff_number = dif.sum()
        diff_p=round(diff_number/num_rows,2)
        test="target"
        # Verifica che il numero di modifiche sia corretto
        self.assertEqual(diff_p,percentage)

    


if __name__ == "__main__":
    unittest.main()



from pucktrick.utils import *
import pandas as pd


def noiseCategoricalStringNewExistingValues(train_df,column,percentage):
  extracted_list=sampleList(percentage,len(train_df[column]))
  noise_df= train_df.copy()
  unique_values = noise_df[column].unique()
  for i, value in enumerate(extracted_list):
     while True:
        new_value=np.random.choice(unique_values)
        if new_value != noise_df.loc[i, column]:
            noise_df.loc[i, column] = new_value
            break
  return noise_df




def noiseCategoricalStringExtendedExistingValues(original_df, train_df,column,percentage):
    noise_df= train_df.copy()
    noise_df['id1'] = range(len(noise_df))
    new_df,newPercentage=generateSubdf(original_df, noise_df,column,percentage)
    if newPercentage==0:
      return train_df
    modified_df= noiseCategoricalStringNewExistingValues(new_df,column,newPercentage)
    noise_df=mergeDataframe(noise_df,modified_df)
    return noise_df

def noiseCategoricalStringNewFakeValues(train_df,column,percentage):
  extracted_list=sampleList(percentage,len(train_df[column]))
  noise_df= train_df.copy()
  unique_values = noise_df[column].unique()
  for i, value in enumerate(extracted_list):
    noise_df.loc[i, column] = ''.join(np.random.choice(list('abcdefghijklmnopqrstuvwxyz'), size=5))
  return noise_df

def noiseCategoricalStringExstendedFakeValues(original_df, train_df,column,percentage):
    noise_df= train_df.copy()
    noise_df['id1'] = range(len(noise_df))
    new_df,newPercentage=generateSubdf(original_df, noise_df,column,percentage)
    if newPercentage==0:
      return train_df
    modified_df= noiseCategoricalStringNewFakeValues(new_df,column,newPercentage)
    noise_df=mergeDataframe(noise_df,modified_df)
    return noise_df

def noiseCategoricalIntNewExistingValues(train_df,column,percentage):
  extracted_list=sampleList(percentage,len(train_df[column]))
  noise_df= train_df.copy()
  unique_values = noise_df[column].unique()
  for i, value in enumerate(extracted_list):
     while True:
        new_value=np.random.choice(unique_values)
        if new_value != noise_df.loc[i, column]:
            noise_df.loc[i, column] = new_value
            break
  return noise_df

def noiseCategoricalIntExtendedExistingValues(original_df, train_df,column,percentage):
    noise_df= train_df.copy()
    noise_df['id1'] = range(len(noise_df))
    new_df,newPercentage=generateSubdf(original_df, noise_df,column,percentage)
    if newPercentage==0:
      return train_df
    modified_df= noiseCategoricalStringNewExistingValues(new_df,column,newPercentage)
    noise_df=mergeDataframe(noise_df,modified_df)
    return noise_df



def noiseDiscreteExtended(original_df, train_df,column,percentage):
    noise_df= train_df.copy()
    noise_df['id1'] = range(len(noise_df))
    new_df,newPercentage=generateSubdf(original_df, noise_df,column,percentage)
    if newPercentage==0:
      return train_df
    modified_df= noiseDiscreteNew(new_df,column,newPercentage)
    noise_df=mergeDataframe(noise_df,modified_df)
    return noise_df


def noiseDiscreteNew(train_df,column,percentage):
    extracted_list=sampleList(percentage,len(train_df[column]))
    noise_df= train_df.copy()
    min = noise_df[column].min()
    max = noise_df[column].max()

    for i, value in enumerate(extracted_list):
       while True:
        new_value=random.randint(min,max)
        if new_value != noise_df.loc[i, column]:
            noise_df.loc[i, column] = new_value
            break
    return noise_df


def noiseBinaryNew(train_df,target,percentage):
  noise_df= train_df.copy()
  extracted_list=sampleList(percentage,len(noise_df[target])-1)
  for i, value in enumerate(extracted_list):
    if pd.isna(noise_df.loc[i, target]):
        noise_df.loc[i, target]=0
    else:
      existingValue=noise_df.loc[i, target]
      noise_df.loc[i, target]=1-existingValue
  return noise_df

def noiseBinaryExtended(original_df, train_df,column,percentage):
    noise_df= train_df.copy()
    noise_df['id1'] = range(len(noise_df))
    new_df,newPercentage=generateSubdf(original_df, noise_df,column,percentage)
    if newPercentage==0:
      return train_df
    modified_df= noiseBinaryNew(new_df,column,newPercentage)
    noise_df=mergeDataframe(noise_df,modified_df)
    return noise_df
def noiseContinueExtended(original_df, train_df,column,percentage):
    noise_df= train_df.copy()
    noise_df['id1'] = range(len(noise_df))
    new_df,newPercentage=generateSubdf(original_df, noise_df,column,percentage)
    if newPercentage==0:
      return train_df
    modified_df= noiseContinueNew(new_df,column,newPercentage)
    noise_df=mergeDataframe(noise_df,modified_df)
    return noise_df


def noiseContinueNew(train_df,column,percentage):
    extracted_list=sampleList(percentage,len(train_df[column]))
    noise_df= train_df.copy()
    min = noise_df[column].min()
    max = noise_df[column].max()

    for i, value in enumerate(extracted_list):
       while True:
        new_value=random.uniform(min,max)
        if new_value != noise_df.loc[i, column]:
            noise_df.loc[i, column] = new_value
            break
    return noise_df





import unittest
from pucktrick.missing import *
from pucktrick.utils import *
import pandas as pd
import numpy as np
from pandas.testing import assert_frame_equal

class Test_outlies(unittest.TestCase):
  def test_MissingAll(self):
      percentage=0.5
      num_rows = 1000
      fake_df=create_fake_table(num_rows)
      target='f3'
      null_values=fake_df[target].isna().sum()
      noise_df=missingNew(fake_df,target,percentage)
      percentage=0.7
      noise_df=missingExtended(fake_df,noise_df,target,percentage)
      new_null_values=noise_df[target].isna().sum()
      if null_values>0:
        diff=round((new_null_values-null_values)/null_values,2)
        self.assertEqual(diff, percentage)
      else:
          df_len=len(noise_df)
          rows=int(df_len*percentage)
          self.assertEqual(new_null_values, rows)


if __name__ == "__main__":
    unittest.main()
